@@ -70,11 +70,12 @@ import errno
 
 from hashlib import md5
 from io import BytesIO, IOBase
-from types import GeneratorType, ModuleType
+from types import GeneratorType, ModuleType, FunctionType, BuiltinFunctionType
 from collections import deque
 from binascii import a2b_hex, b2a_hex
 
-VERSION = '0.15'
+VERSION = '0.16'
+MIN_COMPAT_VERSION = '0.16'
 
 PYTHON3 = sys.version_info >= (3, 0)
 if PYTHON3:
@@ -102,19 +103,18 @@ try:
 except ImportError:
     epoll = None
 
-__all__ = ['AutoThreadingClient', 'Client', 'ClientGroup', 'Connection',
-           'DefaultRemoteProxy', 'RemoteException',
-           'RemoteProxy', 'SelfIterProxy', 'ServerObject',
-           'add_key_options', 'call_method_multi', 'call_multi', 'decode_key',
-           'define_common', 'definitions_module', 'eval_multi', 'exec_return',
-           'execblock_multi', 'generate_key', 'key_from_options',
-           'make_server_class', 'make_singleton', 'read_key_file',
-           'write_key_file', 'parse_connection_spec']
+__all__ = ['AutoThreadingClient', 'ThreadedClient', 'Client', 'ClientGroup',
+           'Connection', 'DefaultRemoteProxy', 'RemoteException', 'RemoteProxy',
+           'SelfIterProxy', 'ServerObject', 'add_key_options',
+           'call_method_multi', 'call_multi', 'decode_key', 'define_common',
+           'definitions_module', 'eval_multi', 'exec_return', 'execblock_multi',
+           'generate_key', 'key_from_options', 'make_server_class',
+           'make_singleton', 'read_key_file', 'write_key_file',
+           'parse_connection_spec'] 
 
 ## Special value for TID which causes thread to exit after one request
 SINGLE = -1
 
-
 _nonblocking_errors = set()
 for k in ('EWOULDBLOCK', 'EINTR', 'EINPROGRESS', 'EAGAIN'):
     val = getattr(errno, k, None)
@@ -122,6 +122,10 @@ for k in ('EWOULDBLOCK', 'EINTR', 'EINPROGRESS', 'EAGAIN'):
         _nonblocking_errors.add(val)
 del k, val
 
+## Methods we do not want to auto-export.
+_forbidden_slots = set(['__init__', '__new__', '__del__', '__getattr__',
+                        '__getattribute__', '__setattr__', '__delattr__'])
+
 HMAC_MOD = md5
 HMAC_SIZE = len(HMAC_MOD().digest())
 ERROR, CHALLENGE, CHALLENGE_RESPONSE, KEEPALIVE, REQUEST, RESPONSE, INIT = range(7)
@@ -196,32 +200,40 @@ def call_method_multi(_objects, _method, *a, **kw):
     resp = [o.raw_call_method(_method, a, kw, nonblocking=True) for o in _objects]
     return [r.wait() for r in resp]
 
-def make_server_class(typ, export_methods=(), export_attrs=(), proxy_class=None):
+def make_server_class(typ, export_methods=('@auto',), export_attrs=('@auto',),
+                      export_attrs_cache=(), proxy_class=None):
     '''Mark a pre-existing class as a server-object class. This function will
     attempt to assign attributes to 'typ'; if that fails, the type will be added
     to a global dictionary.
     
     typ is an old-style or new-style class.
+
     export_methods is a sequence of method names that the class exports.
+
     export_attrs is a sequence of attribute names that the class exports.
+
+    export_attrs_cache is a sequence of attribute names that the class exports,
+    and whose value should be cached on the client.
+
     proxy_class is a class to instantiate as a proxy for `typ`.
 
     export_methods and export_attrs are ignored when proxy_class is given.
     '''
 
+    typedat = _make_typedat(
+        typ, export_methods, export_attrs, export_attrs_cache, proxy_class)
+
     try:
-        typ.export_methods = export_methods
-        typ.export_attrs = export_attrs
-        typ.proxy_class = proxy_class
-        typ._is_server_object = True
+        typ._easycluster_server_object_typedat = typedat
     except Exception:
         # If setting attributes on the type fails, it probably means it's an
         # extension type that doesn't have a __dict__. In that case, store it in
         # _extra_base_classes; Server.persistent_id searches through the MRO of
         # new-style classes to find the first one in this dict.
-        _extra_base_classes[typ] = export_methods, export_attrs, proxy_class
+        _extra_base_classes[typ] = typedat
 
-def make_singleton(obj, export_methods=(), export_attrs=(), proxy_class=None):
+def make_singleton(obj, export_methods=('@auto',), export_attrs=('@auto',), 
+                   export_attrs_cache=(), proxy_class=None):
     '''Mark a singleton object instance as a server object. When this object is
     referenced in a return value, it will be instantiated on the client as a
     simple remote proxy.
@@ -238,7 +250,9 @@ def make_singleton(obj, export_methods=(), export_attrs=(), proxy_class=None):
         ## type(obj) is not weakref-capable.
         wr = None
         
-    _singletons[addr] = export_methods, export_attrs, proxy_class, wr, obj
+    typedat = _make_typedat(
+        obj.__class__, export_methods, export_attrs, export_attrs_cache, proxy_class)
+    _singletons[addr] = typedat, wr, obj
 
 class DummyConnection(object):
     '''Replacement object for _connection in invalidated proxies.'''
@@ -345,7 +359,23 @@ class RemoteProxy(RemoteProxyBase):
         return self._oid
 
     def raw_call_method(self, method, args, kwargs={}, nonblocking=False,
-                        oncomplete=None, onerror=None, threadid=None):
+                        oncomplete=None, onerror=None, threadid=None,
+                        origexc=True):
+        '''Call a method on the remote object. 
+
+        method is the method to call.
+        args is a tuple of positional arguments to the method
+        kwargs is a dict of keyword arguments to the method
+        
+        See Client.send_request for the meaning of the 'nonblocking',
+        'oncomplete', 'onerror', 'threadid', and 'origexc' arguments. 
+        '''
+        return self._connection.send_request(self._oid, method, args, kwargs,
+                                         nonblocking, oncomplete, onerror, threadid, origexc)
+
+    def raw_get_attribute(self, attr, kwargs={}, nonblocking=False,
+                        oncomplete=None, onerror=None, threadid=None,
+                        origexc=True):
         '''Call a method on the remote object. 
 
         method is the method to call.
@@ -353,20 +383,54 @@ class RemoteProxy(RemoteProxyBase):
         kwargs is a dict of keyword arguments to the method
 
         See Client.send_request for the meaning of the 'nonblocking',
-        'oncomplete', 'onerror', and 'threadid' arguments. 
+        'oncomplete', 'onerror', 'threadid', and 'origexc' arguments. 
         '''
-        return self._connection.send_request(self._oid, method, args, kwargs,
-                                         nonblocking, oncomplete, onerror, threadid)
+        return self._connection.send_request(0, 'getattr', (self, name))
+
+    def raw_set_attribute(self, attr, val, kwargs={}, nonblocking=False,
+                        oncomplete=None, onerror=None, threadid=None,
+                        origexc=True):
+        '''Call a method on the remote object. 
+
+        method is the method to call.
+        args is a tuple of positional arguments to the method
+        kwargs is a dict of keyword arguments to the method
+
+        See Client.send_request for the meaning of the 'nonblocking',
+        'oncomplete', 'onerror', 'threadid', and 'origexc' arguments. 
+        '''
+        return self._connection.send_request(0, 'setattr', (self, name, val))
 
     def __repr__(self):
         return '<RemoteProxy for oid %d on %s>' % (self._oid, self._connection.peername)
 
 class DefaultRemoteProxy(RemoteProxy):
-    """A remote proxy for server-side classes that don't explicitly export
-    any methods.
+    """A remote proxy for server-side classes that want to automatically export
+    attributes and methods.
     """
+    def __init__(self, *args, **kw):
+        super(DefaultRemoteProxy, self).__init__(*args, **kw)
+        self._known_methods = {}
+
     def __getattr__(self, attr):
-        return _create_wrapper_method(attr).__get__(self)
+        meth = self._known_methods.get(attr)
+        if meth:
+            return meth.__get__(self)
+        
+        nbr = self._connection.send_request(0, 'getmethod', (self, attr), nonblocking=True)
+        try:
+            ismethod, val = nbr.wait(False)
+            if not ismethod:
+                return val
+            meth = self._known_methods[attr] = _create_wrapper_method(attr)
+            return meth.__get__(self)
+
+        except RemoteException as re:
+            # Don't report AttributeError: we may be called by hasattr, or
+            # getattr(a, 'attr', None).
+            if not isinstance(re.orig, AttributeError):
+                self._connection.report_exception(re)
+            raise re.orig
 
 class SelfIterProxy(RemoteProxy):
     """A proxy class for an iterator which returns itself when __iter__ is
@@ -376,11 +440,15 @@ class SelfIterProxy(RemoteProxy):
     def __iter__(self):
         return self
 
+class _TypeData(str):
+    pass
+
 class ServerObject:
     '''A base class for objects which can be referenced by a client.'''
-    _is_server_object = True
-    export_methods = True
-    export_attrs = ()
+    _easycluster_server_object_typedat = _TypeData('')
+    export_methods = '@auto',
+    export_attrs = '@auto',
+    export_attrs_cache = ()
     proxy_class = None
 
 class RemoteException(Exception):
@@ -399,11 +467,12 @@ class RemoteException(Exception):
 
 class NonblockingResponse:
     '''Returned by remote calls when the `nonblocking` flag is set to True.'''
-    def __init__(self, client, request_id):
+    def __init__(self, client, request_id, origexc):
         self.client = client
         self.request_id = request_id
         self._have_response = False
         self._response_val = None
+        self._origexc = origexc
 
     @property
     def have_response(self):
@@ -425,21 +494,25 @@ class NonblockingResponse:
         self._response_val = val
         self._have_response = True
 
-    def wait(self, origexc=True):
+    def wait(self, origexc=None):
         '''Waits for the request to complete.
         
-        If the call causes an exception and `origexc` is True (default), the
-        remote exception traceback is printed to sys.stderr and the original
-        exception is raised. If origexc is False, a RemoteException is raised
-        instead and the remote traceback is not printed.
+        If the call causes an exception and `origexc` is True, the remote
+        exception traceback is printed to sys.stderr and the original exception
+        is raised. If origexc is False, a RemoteException is raised instead and
+        the remote traceback is not printed.
         '''
 
+        if origexc is None:
+            origexc = self._origexc
+
         self.client._wait_for_response(self)
 
         rval = self._response_val
         if type(rval) is RemoteException:
             if origexc:
-                self.client.report_exception(rval)
+                if origexc != 'quiet':
+                    self.client.report_exception(rval)
                 raise rval.orig
             raise rval
         return rval
@@ -490,12 +563,13 @@ class Connection:
         '''The TCP port passed to connect(), or None for a local connection'''
         return self._remote_port
 
-    def connect(self, host, port=DEFAULT_PORT):
+    def connect(self, host, port=DEFAULT_PORT, peername=None):
         '''Connect to a remote host.'''
         self._remote_host = host
         self._remote_port = port
-        self._peername = '%s:%d' % (host, port)
-        self.set_socket(socket.create_connection((host, port)))
+        if peername is None:
+            peername = '%s:%d' % (host, port)
+        self.set_socket(socket.create_connection((host, port)), peername)
         
     def reconnect(self):
         '''Reconnect to the remote host. This will reset the remote environment,
@@ -503,7 +577,7 @@ class Connection:
         if self._remote_host is None:
             raise ValueError('Cannot reconnect before first connection.')
         self.close()
-        self.connect(self._remote_host, self._remote_port)
+        self.connect(self._remote_host, self._remote_port, self._peername)
 
     def set_socket(self, sock, peername=None):
         '''Use an already-open socket for this connection.'''
@@ -558,6 +632,18 @@ class Connection:
 
     def _got_init(self, p):
         self._peer_version = p.get('version')
+        min_vers_tuple = tuple(int(v) for v in MIN_COMPAT_VERSION.split('.'))
+        peer_vers_tuple = tuple(int(v) for v in self._peer_version.split('.'))
+        if peer_vers_tuple < min_vers_tuple:
+            svr_vers = VERSION
+            client_vers = self._peer_version
+            me = 'server'
+            if isinstance(self, Client):
+                svr_vers, client_vers = client_vers, svr_vers
+                me = 'client'
+
+            raise ValueError('Incompatible EasyCluster version: client has version %s, server has version %s; '
+                             '%s requires at least version %s' % (client_vers, svr_vers, me, MIN_COMPAT_VERSION))
         return True
 
     def _got_keepalive(self):
@@ -638,7 +724,11 @@ class Connection:
                     elif rmsg == RESPONSE:
                         return self._got_response(pick)
                     elif rmsg == INIT:
-                        return self._got_init(pick)
+                        try:
+                            return self._got_init(pick)
+                        except ValueError as e:
+                            self.send_message(ERROR, bytes(str(e)))
+                            raise
 
                 elif rmsg == KEEPALIVE:
                     self._got_keepalive()
@@ -679,11 +769,11 @@ class _GlobalProxy:
         except AttributeError:
             return _GlobalProxy(self._client, self._name + '.' + name)
 
-    def __setattribute__(self, name, newval):
+    def __setattr__(self, name, newval):
         if name == 'value':
             self._client.send_request(0, 'set_global', (self._name, newval))
             return newval
-        return object.__setattribute__(self, name, newval)
+        return object.__setattr__(self, name, newval)
     
     def __call__(self, *args, **kw):
         return self._client.raw_call(self._name, args, kw, *_get_special_keywords(kw))
@@ -843,7 +933,8 @@ class Client(Connection):
         return True
 
     def send_request(self, oid, meth, args, kwargs={}, nonblocking=False,
-                     oncomplete=None, onerror=None, threadid=None):
+                     oncomplete=None, onerror=None, threadid=None,
+                     origexc=True):
         '''Send a request to the remote server. This method is normally not
         called directly.
 
@@ -864,6 +955,12 @@ class Client(Connection):
 
         If nonblocking is True, a NonblockingResponse object is returned instead.
 
+        If origexc is True (default), and the request raises an exception, it
+        will print the remote stack trace to the screen and raise the original
+        exception. If it is False, a RemoteException is raised instead. If it is
+        the value 'quiet', then the original exception is raised without a stack
+        trace being printed.
+
         threadid is the remote thread to queue the response in. If not
         specified, the current default will be used. The default can be changed
         by calling set_default_thread(). If the specified thread does not
@@ -872,7 +969,7 @@ class Client(Connection):
         then exits.
 
         If your client application is multithreaded, and you want to have a
-        separate thread on the server for each client, see AutoThreadingClient.
+        separate thread on the server for each client, see ThreadedClient.
 
         See NonblockingResponse.wait for details on exception handling.
         '''
@@ -907,7 +1004,7 @@ class Client(Connection):
             self._completion[request_id] = oncomplete, onerror
             rval = request_id
         else:
-            rval = NonblockingResponse(self, request_id)
+            rval = NonblockingResponse(self, request_id, origexc)
             self._nonblocking_responses[request_id] = weakref.ref(rval)
         if rq is not None:
             self.send_message(REQUEST, rq)
@@ -940,19 +1037,19 @@ class Client(Connection):
             self.read_response()
 
     def update_definitions(self, nonblocking=False,
-                           oncomplete=None, onerror=None, threadid=None):
+                           oncomplete=None, onerror=None, threadid=None, origexc=True):
         '''Call this after calling define_common to send any new common
         definitions to the server.'''
 
         source = self._defmod._source_code
         current_count = self._definition_count
         rv = self.send_request(0, 'update_definitions', (source[current_count:],), {}, 
-                          nonblocking, oncomplete, onerror, threadid)
+                          nonblocking, oncomplete, onerror, threadid, origexc)
         self._definition_count = len(source)
         return rv
 
     def raw_call(self, func, args, kwargs={}, nonblocking=False,
-                        oncomplete=None, onerror=None, threadid=None):
+                        oncomplete=None, onerror=None, threadid=None, origexc=True):
         '''Call a function on the remote server.
 
         func is the absolute name of the function to call. If 'func'
@@ -963,38 +1060,38 @@ class Client(Connection):
         kwargs is a dict of keyword arguments to the function.
 
         See Client.send_request for the meaning of the 'nonblocking',
-        'oncomplete', 'onerror', and 'threadid' arguments. 
+        'oncomplete', 'onerror', 'threadid', and 'origexc' arguments. 
         '''
         return self.send_request(0, 'call', (func, args, kwargs), {},
-                                 nonblocking, oncomplete, onerror, threadid)
+                                 nonblocking, oncomplete, onerror, threadid, origexc)
         
     def call(self, _func, *args, **kwargs):
         '''Wrapper for raw_call which allows specifying
         arguments directly.'''
         return self.raw_call(_func, args, kwargs, *_get_special_keywords(kwargs))
 
-    def eval(self, expr, nonblocking=False, oncomplete=None, onerror=None, threadid=None):
+    def eval(self, expr, nonblocking=False, oncomplete=None, onerror=None, threadid=None, origexc=True):
         '''Evaluate an expression on the remote server. The code is executed in
         the context of the 'easycluster.remote_code' module.
 
         See Client.send_request for the meaning of the 'nonblocking',
-        'oncomplete', 'onerror', and 'threadid' arguments. 
+        'oncomplete', 'onerror', 'threadid', and 'origexc' arguments. 
         '''
         return self.send_request(0, 'eval', (expr, 'eval'), {},
-                                 nonblocking, oncomplete, onerror, threadid)
+                                 nonblocking, oncomplete, onerror, threadid, origexc)
 
-    def execblock(self, code, nonblocking=False, oncomplete=None, onerror=None, threadid=None):
+    def execblock(self, code, nonblocking=False, oncomplete=None, onerror=None, threadid=None, origexc=True):
         '''Run a block of code on the remote server. The code is executed in the
         context of the 'easycluster.remote_code' module. Returns None unless the
         block executes exec_return().
 
         See Client.send_request for the meaning of the 'nonblocking',
-        'oncomplete', 'onerror', and 'threadid' arguments. 
+        'oncomplete', 'onerror', 'threadid', and 'origexc' arguments. 
         '''
         return self.send_request(0, 'eval', (code, 'exec'), {},
-                                 nonblocking, oncomplete, onerror, threadid)
+                                 nonblocking, oncomplete, onerror, threadid, origexc)
 
-    def import_modules(self, names, nonblocking=False, oncomplete=None, onerror=None, threadid=None):
+    def import_modules(self, names, nonblocking=False, oncomplete=None, onerror=None, threadid=None, origexc=True):
         '''Imports one or more modules on the remote server.
 
         client.import_modules(["sys", "os"])
@@ -1006,18 +1103,18 @@ class Client(Connection):
         `names` is a list of modules to import.
 
         See Client.send_request for the meaning of the 'nonblocking',
-        'oncomplete', 'onerror', and 'threadid' arguments. 
+        'oncomplete', 'onerror', 'threadid', and 'origexc' arguments. 
 
         '''
 
         return self.send_request(0, 'import_modules', (names,), {},
-                                 nonblocking, oncomplete, onerror, threadid)
+                                 nonblocking, oncomplete, onerror, threadid, origexc)
 
-    def import_module(self, name, nonblocking=False, oncomplete=None, onerror=None, threadid=None):
+    def import_module(self, name, nonblocking=False, oncomplete=None, onerror=None, threadid=None, origexc=True):
         '''Imports a module on the remote server. Equivalent to import_modules([name]).'''
 
         return self.send_request(0, 'import_modules', ([name],), {},
-                                 nonblocking, oncomplete, onerror, threadid)
+                                 nonblocking, oncomplete, onerror, threadid, origexc)
 
     @property
     def server_version(self):
@@ -1058,22 +1155,32 @@ class Client(Connection):
             # Check if we have a type created for this object
             if proxclas is None and len(l) > 2:
                 typedat = l[2]
-                if typedat == ':!:':
-                    proxclas = DefaultRemoteProxy
-                elif typedat[0] == ':':
-                    _, export_methods, export_attrs = typedat.split(':')
+                if typedat[0] == ':':
+                    _, export_methods_text, export_attrs_text = typedat.split(':')
+
                     base_class = RemoteProxy
-                    if export_methods == '!':
-                        base_class = DefaultRemoteProxy
-                        export_methods = []
-                    else:
-                        export_methods = export_methods.split(',')
-                    export_attrs = export_attrs.split(',')
+                    export_methods = []
+                    export_attrs = []
+                    export_attrs_cache = []
+
+                    if export_methods_text:
+                        export_methods = export_methods_text.split(',')
+
+                    if export_attrs_text:
+                        for attr in export_attrs_text.split(','):
+                            if attr.startswith('!'):
+                                export_attrs_cache.append(attr[1:])
+                            elif attr == '@auto':
+                                base_class = DefaultRemoteProxy
+                            else:
+                                export_attrs.append(attr)
+                    
                     typedict = {'__module__':__name__, 
                                 'proxy_methods': export_methods,
-                                'proxy_attrs': export_attrs
+                                'proxy_attrs': export_attrs,
+                                'proxy_attrs_cache': export_attrs_cache
                                 }
-                    proxclas = RemoteProxyMeta('dynamic_proxy_' + '_'.join(export_methods + export_attrs), 
+                    proxclas = RemoteProxyMeta('dynamic_proxy_' + '_'.join(export_methods + export_attrs + export_attrs_cache), 
                                                (base_class,), typedict)
 
                 else:
@@ -1172,7 +1279,7 @@ class _ATCThread(threading.Thread):
         finally:
             atc.close(*closeargs)
 
-class AutoThreadingClient(Client, threading.Thread):
+class ThreadedClient(Client, threading.Thread):
     '''A client that automatically creates unique threads on the server for each
     thread on the client that makes remote calls.
 
@@ -1190,7 +1297,7 @@ class AutoThreadingClient(Client, threading.Thread):
     def __init__(self, key, host=None, port=DEFAULT_PORT,
                  definitions='easycluster.remote_code', keep_alive_interval=10, 
                  keep_alive_timeout=60):
-        '''Creates a new AutoThreadingClient. Arguments are the same as Client.'''
+        '''Creates a new ThreadedClient. Arguments are the same as Client.'''
 
 
         ## use pseudo-thread IDs instead of sending real ones
@@ -1231,8 +1338,8 @@ class AutoThreadingClient(Client, threading.Thread):
 
 
     def read_response(self, *args):
-        '''Do not call this method on AutoThreadingClient.'''
-        raise TypeError('read_response should not be called on AutoThreadingClient objects')
+        '''Do not call this method on ThreadedClient.'''
+        raise TypeError('read_response should not be called on ThreadedClient objects')
 
     ## called with lock held
     def _thread_id(self):
@@ -1347,6 +1454,9 @@ class ClientGroup:
                             if max == 0:
                                 return
 
+## Backwards-compatible alias
+AutoThreadingClient = ThreadedClient
+
 def exec_return(rval):
     '''When called from within code sent to Client.eval, causes eval to return 'rval'.'''
     raise _ExecReturn(rval)
@@ -1363,11 +1473,12 @@ def definitions_module(name, keep_source=True):
     mod = ModuleType(name, 'EasyCluster common definitions')
     sys.modules[name] = mod
     mod._source_code = source = []
-    def define(code):
+    def define(code, local=True):
         if not code.endswith('\n'):
             code += '\n'
         cb = compile(code, '<easycluster definitions %r>' % name, 'exec', dont_inherit=True)
-        eval(cb, mod.__dict__)
+        if local:
+            eval(cb, mod.__dict__)
         if keep_source:
             source.append(code)
             
@@ -1377,11 +1488,11 @@ def definitions_module(name, keep_source=True):
     del source[:]
     return mod
 
-def define_common(code, name='easycluster.remote_code'):
+def define_common(code, name='easycluster.remote_code', local=True):
     '''Define common classes in the specified module. Equivalent to:
-    definitions_module(name).define(code)
+    definitions_module(name).define(code, local)
     '''
-    definitions_module(name).define(code)
+    definitions_module(name).define(code, local)
 
 def decode_key(key):
     '''If 'key' is a hex-encoded string of the proper size, returns the actual key,
@@ -1437,8 +1548,6 @@ def key_from_options(opts, warn=True):
         if warn:
             print('SECURITY WARNING: using key from command line.', file=sys.stderr)
         key = decode_key(opts.key)
-    if key == b'' and warn:
-        print('SECURITY WARNING: running with blank key.', file=sys.stderr)
     return key
 
 def parse_connection_spec(spec, default_port=DEFAULT_PORT, default_key=b'', custom_options=(), warn=True):
@@ -1463,7 +1572,7 @@ def parse_connection_spec(spec, default_port=DEFAULT_PORT, default_key=b'', cust
     options, or True to allow any arbitrary options.
 
     The return value is a dictionary of parameters which can be passed to the
-    Client or AutoThreadingClient constructors:
+    Client or ThreadedClient constructors:
 
      >>> params = easycluster.parse_connection_spec(sys.argv[1])
      >>> client = easycluster.Client(**params)
@@ -1565,6 +1674,79 @@ class _ExecReturn(BaseException):
 # Internal functions
 ###############################
 
+def _get_typedat(obj, typ):
+    '''Returns the saved type data for object `obj` of type `typ`.'''
+    sdat = _singletons.get(id(obj))
+    if sdat is not None:
+        typedat, wr, sobj = sdat
+        if sobj is not None or (wr is not None and wr() is obj):
+            return typedat
+    
+    typedat = getattr(typ, '_easycluster_server_object_typedat', None)
+    if type(typedat) == _TypeData:
+        if typedat == '':
+            typedat = _make_typedat(
+                typ, typ.export_methods, typ.export_attrs, 
+                typ.export_attrs_cache, typ.proxy_class)
+            typ._easycluster_server_object_typedat = typedat
+        return typedat
+
+    ebc = _extra_base_classes.get
+    for b in getattr(typ, '__mro__', ()):
+        r = ebc(b)
+        if r:
+            return r
+    return None
+
+def _fill_old_style_mro(typ, mro):
+    if typ not in mro:
+        mro.append(typ)
+    for b in typ.__bases__:
+        _fill_old_style_mro(b, mro)
+
+def _class_mro(typ):
+    mro = getattr(typ, '__mro__', None)
+    if mro is not None:
+        return mro[:-1]
+    mro = []
+    _fill_old_style_mro(typ, mro)
+    return mro
+
+def _make_typedat(typ, export_methods, export_attrs, export_attrs_cache, proxy_class):
+    '''
+    Generate a type data string for the given set of exported methods and
+    attributes or proxy class.
+    '''
+    if isinstance(export_methods, str):
+        export_methods = (export_methods,)
+    if isinstance(export_attrs, str):
+        export_attrs = (export_attrs,)
+    if isinstance(export_attrs_cache, str):
+        export_attrs_cache = (export_attrs_cache,)
+
+    if proxy_class:
+        return _TypeData(proxy_class.__module__ + '.' + proxy_class.__name__)
+    else:
+        if '@auto' in export_methods:
+            export_methods = set(export_methods)
+            export_methods.discard('@auto')
+            attrs = {}
+            for cls in _class_mro(typ):
+                if cls is not object:
+                    attrs.update(cls.__dict__)
+
+            functypes = (FunctionType, BuiltinFunctionType, staticmethod, classmethod)
+            export_methods = [name for name, val in attrs.items()
+                              if (name not in _forbidden_slots and 
+                                  isinstance(val, functypes))]
+
+        export_attrs = set(export_attrs or ())
+        for cattr in (export_attrs_cache or ()):
+            export_attrs.discard(cattr)
+            export_attrs.add('!' + cattr)
+
+        return _TypeData(':' + ','.join(sorted(export_methods or ())) + ':' + ','.join(sorted(export_attrs)))
+
 def _get_module(name):
     '''Imports and returns a module object.'''
     components = name.split('.')
@@ -1605,27 +1787,26 @@ def _get_special_keywords(kwargs):
     oncomplete = kwargs.pop('oncomplete', None)
     onerror = kwargs.pop('onerror', None)
     threadid = kwargs.pop('threadid', None)
-    return nonblocking, oncomplete, onerror, threadid
+    origexc = kwargs.pop('origexc', True)
+    return nonblocking, oncomplete, onerror, threadid, origexc
 
 def _add_extra_classes():
-    _file_export = ((), (), FileProxy)
-
-    _extra_base_classes[GeneratorType] = ((), (), GeneratorProxy)
-    _extra_base_classes[IOBase] = _file_export
+    make_server_class(GeneratorType, proxy_class=GeneratorProxy)
+    make_server_class(IOBase, proxy_class=FileProxy)
     try:
-        _extra_base_classes[file] = _file_export
+        make_server_class(file, proxy_class=FileProxy)
     except NameError:
         pass
 
     try:
         import _io
-        _extra_base_classes[_io._IOBase] = _file_export
+        make_server_class(_io._IOBase, proxy_class=FileProxy)
         del _io
     except (NameError, ImportError):
         pass
     
     # Proxy module objects.
-    _extra_base_classes[ModuleType] = ((), (), None)
+    make_server_class(ModuleType)
     
 _add_extra_classes()
 del _add_extra_classes
