@@ -19,11 +19,14 @@
 # TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
 # SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 
+from __future__ import print_function
+
 import sys
 import os
 import signal
 import subprocess
 import traceback
+import time
 
 NAME = 'EasyCluster'
 MODULE = 'easycluster'
@@ -75,9 +78,7 @@ def install_service():
         else:
             break
     else:
-        print >>sys.stderr, "No method of installing initscripts found (tried Debian, Redhat, LSB)"
-        return 1
-    return 0
+        raise OSError("No method of installing initscripts found (tried Debian, Redhat, LSB)")
 
 def uninstall_service():
     for prog in (['update-rc.d', '-f', MODULE, 'remove'],
@@ -90,8 +91,7 @@ def uninstall_service():
         else:
             break
     else:
-        print >>sys.stderr, "No method of uninstalling initscripts found (tried Debian, Redhat, LSB)"
-        return 1
+        raise OSError("No method of uninstalling initscripts found (tried Debian, Redhat, LSB)")
 
     try:
         os.unlink(INIT_PATH)
@@ -111,7 +111,17 @@ def query_service_installed():
 
 def is_daemon(pid, name):
     try:
-        cmdline = file('/proc/%d/cmdline' % pid, 'r').read().split('\0')
+        os.kill(pid, 0)
+    except OSError:
+        return False
+
+    try:
+        cmdf = '/proc/%d/cmdline' % pid
+        if not os.path.exists(cmdf):
+            ## Proc not mounted or not Linux - just assume pidfile is valid
+            return True
+
+        cmdline = file(cmdf, 'r').read().split('\0')
         if cmdline[1] == '-m' and cmdline[2] == name:
             return True
     except Exception:
@@ -141,34 +151,34 @@ def getpid():
 def start_daemon():
     opid = getpid()
     if opid:
-        print "%s is already running." % NAME
+        print("%s is already running." % NAME)
         return 1
 
-    print "Starting %s:" % NAME,
+    print("Starting %s: " % NAME, end='')
     exitstat = subprocess.call([sys.executable, '-m', MAIN_MODULE, '-S', '-k', KEY_PATH, '-d', '-P', PIDFILE])
     if exitstat:
-        print 'failed.'
+        print('failed.')
         return 2
-    print "ok."
+    print("ok.")
     return 0
 
 def stop_daemon():
     pid = getpid()
     if not pid:
-        print "%s is not running." % NAME
+        print("%s is not running." % NAME)
         return 1
     try:
-        print "Stopping %s:" % NAME,
+        print("Stopping %s: " % NAME, end='')
         
         os.kill(pid, signal.SIGTERM)
         try:
             os.unlink(PIDFILE)
         except Exception:
             pass
-        print "success."
+        print("success.")
         return 0
     except Exception:
-        print "failed."
+        print("failed.")
         return 2
         
 
@@ -188,15 +198,15 @@ def init_main():
             os.kill(pid, signal.SIGHUP)
             exitstat = 1
         else:
-            print "%s is not running." % NAME
+            print("%s is not running." % NAME)
 
     elif cmd == "status":
         pid = getpid()
         if pid:
-            print "%s is running." % NAME
+            print("%s is running." % NAME)
         else:
             exitstat = 1
-            print "%s is not running." % NAME
+            print("%s is not running." % NAME)
 
     elif cmd == "restart":
         stop_daemon()
@@ -204,7 +214,7 @@ def init_main():
         exitstat = start_daemon()
     else:
 
-        print >>sys.stderr, "Usage: %s {start|stop|restart|reload|force-reload}" % sys.argv[0]
+        print("Usage: %s {start|stop|restart|reload|force-reload}" % sys.argv[0], file=sys.stderr)
         exitstat = 1
     return exitstat
 
