@@ -23,6 +23,15 @@
 from __future__ import print_function
 
 import sys
+
+# A little bit of magic to allow running python -m easycluster.server. This
+# allows the user to use '-m' with Python 2.6, which fails when '-m' is used on
+# a package.
+_is_main = __name__ == '__main__'
+if _is_main:
+    __name__ = 'easycluster.server'
+    sys.modules[__name__] = sys.modules['__main__']
+
 import os
 import traceback
 import time
@@ -31,8 +40,10 @@ import errno
 import threading
 import select
 import easycluster as _core
+from weakref import WeakKeyDictionary
 from collections import deque
-from types import ModuleType
+from types import ModuleType, BuiltinMethodType, MethodType, FunctionType
+
 
 ## Stolen from multiprocessing. We can't just use multiprocessing -- it has way
 ## too many weird bugs on Windows.
@@ -138,26 +149,7 @@ def _set_global(name, defmod, value):
         gv = _get_global('.'.join(components[:-1]), defmod)
         setattr(gv, components[-1], value)
 
-def _get_proxy_data(obj, typ):
-    '''Returns the proxy class or exported attributes and methods for object
-    `obj` of type `typ`.
-
-    '''
-    sdat = _core._singletons.get(id(obj))
-    if sdat is not None:
-        export_methods, export_attrs, proxy_class, wr, sobj = sdat
-        if sobj is not None or (wr is not None and wr() is obj):
-            return export_methods, export_attrs, proxy_class
-    
-    if hasattr(typ, '_is_server_object'):
-        return typ.export_methods, typ.export_attrs, typ.proxy_class
-    else:
-        ebc = _core._extra_base_classes
-        for b in getattr(typ, '__mro__', ()):
-            r = ebc.get(b)
-            if r:
-                return r
-    return None, None, None
+_method_types = BuiltinMethodType, MethodType, FunctionType
 
 class RootObject(object):
     '''Utility class instantiated automatically on the server as OID 0.'''
@@ -188,6 +180,12 @@ class RootObject(object):
     def getattr(self, obj, attr):
         return getattr(obj, attr)
 
+    def getmethod(self, obj, attr):
+        val = getattr(obj, attr)
+        if isinstance(val, _method_types):
+            return True, None
+        return False, val
+
     def setattr(self, obj, attr, val):
         return setattr(obj, attr, val)
 
@@ -236,8 +234,8 @@ class Server(_core.Connection):
         self._root = root
         self._local_object_by_id = {0: root}
         self._local_id_by_object = {id(root): 0}
-        self._local_type_by_id = {}
-        self._local_id_by_type = {}
+        #self._local_typedat_by_id = {}
+        self._local_id_by_typedat = {}
         
         self._pending_responses = deque()
         self._keep_alive_interval = None
@@ -361,11 +359,11 @@ class Server(_core.Connection):
 
         ## Find out if the object is a server object or not
         typ = obj.__class__
-        tid = self._local_id_by_type.get(typ)
-        if tid is None:
-            export_methods, export_attrs, proxy_class = _get_proxy_data(obj, typ)
-            if export_methods is None:
-                return None
+        typedat = _core._get_typedat(obj, typ)
+        if typedat is None:
+            return None
+
+        tid = self._local_id_by_typedat.get(typ)
         oid = self._next_object_id
         self._next_object_id = oid + 1
         self._local_id_by_object[id(obj)] = oid
@@ -373,20 +371,13 @@ class Server(_core.Connection):
 
         if tid is not None:
             return '%d\n%d' % (oid, tid)
-        elif not (proxy_class or export_methods or export_attrs):
-            return str(oid)
-        else:
-            tid = self._next_type_id
-            self._next_type_id = tid + 1
-            self._local_id_by_type[typ] = tid
-            self._local_type_by_id[tid] = typ
-            if proxy_class:
-                typedat = proxy_class.__module__ + '.' + proxy_class.__name__
-            else:
-                if export_methods is True:
-                    export_methods = ('!',)
-                typedat = ':' + ','.join(export_methods) + ':' + ','.join(export_attrs)
-            return '%d\n%d\n%s' % (oid, tid, typedat)
+
+        tid = self._next_type_id
+        self._next_type_id = tid + 1
+        self._local_id_by_typedat[typedat] = tid
+        #self._local_typedat_by_id[tid] = typ
+        #print('typedat for %s (%d) is %r' % (typ, tid, typedat))
+        return '%d\n%d\n%s' % (oid, tid, typedat)
 
 class QuietServer(Server):
     '''A subclass of Server which does not print anything to output.'''
@@ -507,6 +498,8 @@ def server_main(args=None):
 
     elif opts.serve:
         key = _core.key_from_options(opts)
+        if key == b'' and warn:
+            print('SECURITY WARNING: running with blank key.', file=sys.stderr)
         if '.' in opts.svrclass:
             svrmod, svrclassname = opts.svrclass.rsplit('.')
             clas = getattr(_core._get_module(svrmod), svrclassname)
@@ -754,3 +747,5 @@ else:
 
         os.setsid()
 
+if _is_main:
+    sys.exit(server_main())
