@@ -24,6 +24,7 @@ import os
 import traceback
 import win32serviceutil
 import win32service
+import pywintypes
 
 sysroot = os.environ.get('SYSTEMROOT', r'c:\windows')
 sysdir = os.path.join(sysroot, 'sysnative')
@@ -74,6 +75,20 @@ def start_service():
 def stop_service():
     return _service_action('stop')
 
+def query_service_installed():
+    hscm = win32service.OpenSCManager(None, None, win32service.SC_MANAGER_ALL_ACCESS)
+    try:
+
+        try:
+            hs = win32serviceutil.SmartOpenService(hscm, EasyClusterService._svc_name_, win32service.SERVICE_ALL_ACCESS)
+            win32service.CloseServiceHandle(hs)
+            return True
+        except pywintypes.error:
+            return False
+    finally:
+        win32service.CloseServiceHandle(hscm)
+    
+
 daemonize = None
 
 if __name__ == '__main__':
