@@ -16,8 +16,8 @@ from __future__ import print_function
 
 import time
 class TestObject(ServerObject):
-    export_methods = 'setval', 'getval'
-    export_attrs = 'val',
+    #export_methods = 'setval', 'getval'
+    #export_attrs = 'val',
 
     def __init__(self, name, val):
         self.name = name
@@ -34,8 +34,8 @@ class TestObject(ServerObject):
         print('TestObject %s: deleted' % self.name)
 
 class TestObject2(ServerObject):
-    export_methods = '__add__',
-    export_attrs = 'val',
+    export_methods = '__add__', 'set_val'
+    export_attrs_cache = 'val',
 
     def __init__(self, val):
         self.val = val
@@ -46,6 +46,9 @@ class TestObject2(ServerObject):
         print('Testobject __add__: %s + %s = %s' % (self.val, oval, rval))
         return TestObject2(rval)
 
+    def set_val(self, val):
+        self.val = val
+
 class TestOldStyleClass:
     def __init__(self, val):
         self.val = val
@@ -82,7 +85,7 @@ class TestObject3(ServerObject):
         oval = o.val
         rval = self.val + oval
         print('Testobject __add__: %s + %s = %s' % (self.val, oval, rval))
-        return TestObject2(rval)
+        return TestObject3(rval)
 
 testobjects = [TestObject('s%d' % i, i) for i in range(3)]
 
@@ -212,6 +215,13 @@ def overload():
     c = a + b
     print('final value = %s' % c.val)
 
+@define_test('Test client-side attribute caching')
+def test_cache():
+    a = l1.TestObject2(3)
+    print('initial value = %s' % a.val)
+    a.set_val(4)
+    assert a.val == 3
+
 @define_test('Test custom proxy classes')
 def custom_proxy():
     a = l1.TestObject3(3)
@@ -314,6 +324,7 @@ def testlocal():
     reftest()
     methods()
     attrs()
+    l1.testobjects.value = None
 
 define_test('Parallel test with local and remote')(parallel3)
 
